import pytest

from rapt2.rapt import Rapt
from rapt2.transformers.qtree.constants import *
from rapt2.treebrd.grammars.dependency_grammar import DependencyGrammar
from rapt2.treebrd.grammars.extended_grammar import ExtendedGrammar
from rapt2.treebrd.treebrd import TreeBRD
from tests.transformers.test_transfomer import TestTransformer


class TestQTreeTransformer(TestTransformer):
    def setUp(self):
        self.translate = self.translate_func(Rapt(grammar="Extended Grammar").to_qtree)


class TestQTreeDependencyTransformer(TestTransformer):
    def setUp(self):
        self.rapt = Rapt(grammar="Dependency Grammar")
        self.translate = self.translate_func(self.rapt.to_qtree)


class TestRelation(TestQTreeTransformer):
    def test_single_relation(self):
        ra = "alpha;"
        expected = [r"\Tree[.$alpha$ ]"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_multiple_relations(self):
        ra = "alpha; beta;"
        expected = [r"\Tree[.$alpha$ ]", r"\Tree[.$beta$ ]"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestSelect(TestQTreeTransformer):
    def test_simple(self):
        ra = r"\select_{a1=a2} alpha;"
        expected = [r"\Tree[.${}_{{(a1 \eq a2)}}$ [.$alpha$ ] ]".format(SELECT_OP)]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestProject(TestQTreeTransformer):
    def test_simple(self):
        ra = r"\project_{a1, a2, a3} alpha;"
        expected = [r"\Tree[.${}_{{a1,\,a2,\,a3}}$ [.$alpha$ ] ]".format(PROJECT_OP)]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestRename(TestQTreeTransformer):
    def test_rename_relation(self):
        ra = r"\rename_{Apex} alpha;"
        expected = [r"\Tree[.${}_{{apex(a1,\,a2,\,a3)}}$ [.$alpha$ ] ]".format(RENAME_OP)]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_rename_attributes(self):
        ra = r"\rename_{(a, b, c)} alpha;"
        expected = [r"\Tree[.${}_{{alpha(a,\,b,\,c)}}$ [.$alpha$ ] ]".format(RENAME_OP)]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_rename_relation_and_attributes(self):
        ra = r"\rename_{apex(a, b, c)} alpha;"
        expected = [r"\Tree[.${}_{{apex(a,\,b,\,c)}}$ [.$alpha$ ] ]".format(RENAME_OP)]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestAssignment(TestQTreeTransformer):
    def test_relation(self):
        ra = "new_alpha := alpha;"
        expected = [r"\Tree[.$new\_alpha(a1,\,a2,\,a3)$ [.$alpha$ ] ]"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_relation_with_attributes(self):
        ra = "new_alpha(a, b, c) := alpha;"
        expected = [r"\Tree[.$new\_alpha(a,\,b,\,c)$ [.$alpha$ ] ]"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestDefinition(TestQTreeTransformer):
    def test_single_attribute(self):
        ra = "R(a);"
        expected = [r"\Tree[.$r(a)$ ]"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_multiple_attributes(self):
        ra = "R(a, b, c);"
        expected = [r"\Tree[.$r(a,\,b,\,c)$ ]"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_two_attributes(self):
        ra = "R(a, b);"
        expected = [r"\Tree[.$r(a,\,b)$ ]"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_multiple_definitions(self):
        ra = "R1(a, b); R2(c, d);"
        expected = [
            r"\Tree[.$r1(a,\,b)$ ]",
            r"\Tree[.$r2(c,\,d)$ ]"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_definition_with_spacing(self):
        ra = "R ( a , b , c ) ;"
        expected = [r"\Tree[.$r(a,\,b,\,c)$ ]"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestJoin(TestQTreeTransformer):
    def test_relation(self):
        ra = r"alpha \join beta;"
        expected = [r"\Tree[.${}$ [.$alpha$ ] [.$beta$ ] ]".format(JOIN_OP)]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestNaturalJoin(TestQTreeTransformer):
    grammar = ExtendedGrammar

    def test_relation(self):
        ra = r"alpha \natural_join beta;"
        expected = [r"\Tree[.${}$ [.$alpha$ ] [.$beta$ ] ]".format(NATURAL_JOIN_OP)]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestThetaJoin(TestQTreeTransformer):
    grammar = ExtendedGrammar

    def test_relation(self):
        ra = r"alpha \join_{a1 = b1} beta;"
        expected = [
            r"\Tree[.${}_{{(a1 \eq b1)}}$ [.$alpha$ ] [.$beta$ ] ]".format(
                THETA_JOIN_OP
            )
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestFullOuterJoin(TestQTreeTransformer):
    grammar = ExtendedGrammar

    def test_relation(self):
        ra = r"alpha \full_outer_join_{a1 = b1} beta;"
        expected = [
            r"\Tree[.${}_{{(a1 \eq b1)}}$ [.$alpha$ ] [.$beta$ ] ]".format(
                FULL_OUTER_JOIN_OP
            )
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_complex_condition(self):
        ra = r"alpha \full_outer_join_{alpha.a1=beta.b1 and b3>50} beta;"
        expected = [
            r"\Tree[.${}_{{((alpha.a1 \eq beta.b1) \land (b3 \gt 50))}}$ [.$alpha$ ] [.$beta$ ] ]".format(
                FULL_OUTER_JOIN_OP
            )
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestLeftOuterJoin(TestQTreeTransformer):
    grammar = ExtendedGrammar

    def test_relation(self):
        ra = r"alpha \left_outer_join_{a1 = b1} beta;"
        expected = [
            r"\Tree[.${}_{{(a1 \eq b1)}}$ [.$alpha$ ] [.$beta$ ] ]".format(
                LEFT_OUTER_JOIN_OP
            )
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_complex_condition(self):
        ra = r"alpha \left_outer_join_{alpha.a1=beta.b1 and b3>50} beta;"
        expected = [
            r"\Tree[.${}_{{((alpha.a1 \eq beta.b1) \land (b3 \gt 50))}}$ [.$alpha$ ] [.$beta$ ] ]".format(
                LEFT_OUTER_JOIN_OP
            )
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestRightOuterJoin(TestQTreeTransformer):
    grammar = ExtendedGrammar

    def test_relation(self):
        ra = r"alpha \right_outer_join_{a1 = b1} beta;"
        expected = [
            r"\Tree[.${}_{{(a1 \eq b1)}}$ [.$alpha$ ] [.$beta$ ] ]".format(
                RIGHT_OUTER_JOIN_OP
            )
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_complex_condition(self):
        ra = r"alpha \right_outer_join_{alpha.a1=beta.b1 and b3>50} beta;"
        expected = [
            r"\Tree[.${}_{{((alpha.a1 \eq beta.b1) \land (b3 \gt 50))}}$ [.$alpha$ ] [.$beta$ ] ]".format(
                RIGHT_OUTER_JOIN_OP
            )
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestUnion(TestQTreeTransformer):
    def test_simple(self):
        ra = r"gamma \union gammatwin;"
        expected = [r"\Tree[.${}$ [.$gamma$ ] [.$gammatwin$ ] ]".format(UNION_OP)]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestIntersect(TestQTreeTransformer):
    grammar = ExtendedGrammar

    def test_simple(self):
        ra = r"gamma \intersect gammatwin;"
        expected = [r"\Tree[.${}$ [.$gamma$ ] [.$gammatwin$ ] ]".format(INTERSECT_OP)]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestDifference(TestQTreeTransformer):
    def test_simple(self):
        ra = r"gamma \difference gammatwin;"
        expected = [r"\Tree[.${}$ [.$gamma$ ] [.$gammatwin$ ] ]".format(DIFFERENCE_OP)]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestFunctionalDependency(TestQTreeDependencyTransformer):
    def setUp(self):
        """Set up test fixtures with proper schema."""
        super().setUp()
        # Override the schema to include the attributes we need for functional dependencies
        self.schema = {
            "alpha": ["a", "b", "c"],
            "beta": ["c", "d", "e"],
            "gamma": ["id", "name", "value"],
        }
        # Update the translate function to use our custom schema
        self.translate = self.translate_func(self.rapt.to_qtree, schema=self.schema)

    def test_simple_functional_dependency(self):
        """Test simple functional dependency without select condition."""
        ra = "fd_{a, b} alpha;"
        expected = [r"\Tree[.$alpha : a \rightarrow b$ ]"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_functional_dependency_with_simple_select(self):
        """Test functional dependency with simple select condition."""
        ra = "fd_{a, b} \\select_{a = 1} alpha;"
        expected = [r"\Tree[.$\sigma_{(a \eq 1)} (alpha) : a \rightarrow b$ ]"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_functional_dependency_with_complex_select(self):
        """Test functional dependency with complex select condition."""
        ra = "fd_{a, b} \\select_{a = 1 and b > 0} alpha;"
        expected = [
            r"\Tree[.$\sigma_{((a \eq 1) \land (b \gt 0))} (alpha) : a \rightarrow b$ ]"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_functional_dependency_with_different_attributes(self):
        """Test functional dependency with different attribute names."""
        ra = "fd_{id, name} \\select_{id > 0} gamma;"
        expected = [r"\Tree[.$\sigma_{(id \gt 0)} (gamma) : id \rightarrow name$ ]"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_functional_dependency_with_or_condition(self):
        """Test functional dependency with OR condition in select."""
        ra = "fd_{a, b} \\select_{a = 1 or a = 2} alpha;"
        expected = [
            r"\Tree[.$\sigma_{((a \eq 1) \lor (a \eq 2))} (alpha) : a \rightarrow b$ ]"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_functional_dependency_with_not_condition(self):
        """Test functional dependency with NOT condition in select."""
        ra = "fd_{a, b} \\select_{not(a = 1)} alpha;"
        expected = [r"\Tree[.$\sigma_{\neg (a \eq 1)} (alpha) : a \rightarrow b$ ]"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_multiple_functional_dependencies(self):
        """Test multiple functional dependencies in one statement."""
        ra = "fd_{a, b} \\select_{a = 1} alpha; fd_{c, d} \\select_{c > 0} beta;"
        expected = [
            r"\Tree[.$\sigma_{(a \eq 1)} (alpha) : a \rightarrow b$ ]",
            r"\Tree[.$\sigma_{(c \gt 0)} (beta) : c \rightarrow d$ ]",
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestDependencyLatexTranslation:
    """Test cases for LaTeX translation of dependency grammar statements."""

    def setup_method(self):
        """Set up test fixtures."""
        self.rapt = Rapt(grammar="Dependency Grammar")
        self.schema = {"R": ["a", "b", "c"], "S": ["x", "y"], "T": ["id", "name"]}

    def test_primary_key_latex_translation(self):
        """Test LaTeX translation of primary key dependencies."""
        # Test single attribute primary key
        latex = self.rapt.to_qtree("pk_{a} R;", self.schema)
        assert len(latex) == 1
        assert "\\text{PK}(r) \\eq {a}" in latex[0]
        assert "\\Tree" in latex[0]

        # Test multiple attributes primary key
        latex = self.rapt.to_qtree("pk_{a, b} R;", self.schema)
        assert len(latex) == 1
        assert "\\text{PK}(r) \\eq {a,\\,b}" in latex[0]

    def test_multivalued_dependency_latex_translation(self):
        """Test LaTeX translation of multivalued dependencies."""
        # Test simple multivalued dependency
        latex = self.rapt.to_qtree("mvd_{a, b} R;", self.schema)
        assert len(latex) == 1
        assert "r : a \\twoheadrightarrow b" in latex[0]
        assert "\\Tree" in latex[0]

    def test_multivalued_dependency_with_select_latex_translation(self):
        """Test LaTeX translation of multivalued dependencies with select conditions."""
        latex = self.rapt.to_qtree("mvd_{a, b} \\select_{a = 1} R;", self.schema)
        assert len(latex) == 1
        assert "\\sigma_{(a \\eq 1)} (r) : a \\twoheadrightarrow b" in latex[0]
        assert "\\Tree" in latex[0]

    def test_inclusion_equivalence_latex_translation(self):
        """Test LaTeX translation of inclusion equivalence dependencies."""
        # Test simple inclusion equivalence
        latex = self.rapt.to_qtree("inc=_{a, b} (R, S);", self.schema)
        assert len(latex) == 1
        assert "r[a] \\equiv s[b]" in latex[0]
        assert "\\Tree" in latex[0]

    def test_inclusion_equivalence_with_select_latex_translation(self):
        """Test LaTeX translation of inclusion equivalence dependencies with select clauses."""
        # Test with select on left
        latex = self.rapt.to_qtree("inc=_{a, b} (\\select_{a = 1} R, S);", self.schema)
        assert len(latex) == 1
        assert "\\sigma_{(a \\eq 1)} (r)[a] \\equiv s[b]" in latex[0]
        assert "\\Tree" in latex[0]

        # Test with select on both sides
        latex = self.rapt.to_qtree("inc=_{a, b} (\\select_{a = 1} R, \\select_{x > 0} S);", self.schema)
        assert len(latex) == 1
        assert "\\sigma_{(a \\eq 1)} (r)[a] \\equiv \\sigma_{(x \\gt 0)} (s)[b]" in latex[0]
        assert "\\Tree" in latex[0]

    def test_inclusion_subsumption_latex_translation(self):
        """Test LaTeX translation of inclusion subsumption dependencies."""
        # Test simple inclusion subsumption
        latex = self.rapt.to_qtree("inc⊆_{a, b} (R, S);", self.schema)
        assert len(latex) == 1
        assert "r[a] \\subseteq s[b]" in latex[0]
        assert "\\Tree" in latex[0]

    def test_inclusion_subsumption_with_select_latex_translation(self):
        """Test LaTeX translation of inclusion subsumption dependencies with select clauses."""
        # Test with select on left
        latex = self.rapt.to_qtree("inc⊆_{a, b} (\\select_{a = 1} R, S);", self.schema)
        assert len(latex) == 1
        assert "\\sigma_{(a \\eq 1)} (r)[a] \\subseteq s[b]" in latex[0]
        assert "\\Tree" in latex[0]

    def test_multiple_dependency_statements_latex_translation(self):
        """Test LaTeX translation of multiple dependency statements."""
        latex = self.rapt.to_qtree(
            "pk_{a} R; mvd_{x, y} S; inc=_{a, b} (R, S); inc⊆_{x, y} (S, T);",
            self.schema,
        )
        assert len(latex) == 4

        # Check first statement (primary key)
        assert "\\text{PK}(r) \\eq {a}" in latex[0]

        # Check second statement (multivalued dependency)
        assert "s : x \\twoheadrightarrow y" in latex[1]

        # Check third statement (inclusion equivalence)
        assert "r[a] \\equiv s[b]" in latex[2]

        # Check fourth statement (inclusion subsumption)
        assert "s[x] \\subseteq t[y]" in latex[3]

    def test_latex_structure_consistency(self):
        """Test that LaTeX output has consistent structure."""
        test_cases = [
            "pk_{a, b} R;",
            "mvd_{x, y} S;",
            "inc=_{a, b} (R, S);",
            "inc⊆_{x, y} (S, T);",
        ]

        for test_case in test_cases:
            latex = self.rapt.to_qtree(test_case, self.schema)
            assert len(latex) == 1
            assert latex[0].startswith("\\Tree")
            assert latex[0].endswith(" ]")
            # Check for appropriate operators based on dependency type
            if "inc=" in test_case or "inc⊆" in test_case:
                assert "\\equiv" in latex[0] or "\\subseteq" in latex[0]
            elif "mvd" in test_case:
                assert "\\twoheadrightarrow" in latex[0]
            else:
                assert "\\text{" in latex[0]

    def test_latex_operators_mapping(self):
        """Test that correct LaTeX operators are used for each dependency type."""
        test_mappings = [
            ("pk_{a} R;", "\\text{PK}"),
            ("mvd_{a, b} R;", "\\twoheadrightarrow"),
            ("inc=_{a, b} (R, S);", "\\equiv"),
            ("inc⊆_{a, b} (R, S);", "\\subseteq"),
        ]

        for test_case, expected_operator in test_mappings:
            latex = self.rapt.to_qtree(test_case, self.schema)
            assert expected_operator in latex[0]

    def test_latex_attributes_formatting(self):
        """Test that attributes are properly formatted in LaTeX output."""
        # Test single attribute
        latex = self.rapt.to_qtree("pk_{a} R;", self.schema)
        assert "\\eq {a}" in latex[0]

        # Test multiple attributes
        latex = self.rapt.to_qtree("pk_{a, b, c} R;", self.schema)
        assert "\\eq {a,\\,b,\\,c}" in latex[0]

    def test_latex_relation_names_formatting(self):
        """Test that relation names are properly formatted in LaTeX output."""
        # Test single relation
        latex = self.rapt.to_qtree("pk_{a} R;", self.schema)
        assert "(r)" in latex[0]

        # Test inclusion dependencies with new format
        latex = self.rapt.to_qtree("inc=_{a, b} (R, S);", self.schema)
        assert "r[a]" in latex[0] and "s[b]" in latex[0]

        latex = self.rapt.to_qtree("inc⊆_{x, y} (S, T);", self.schema)
        assert "s[x]" in latex[0] and "t[y]" in latex[0]

    def test_latex_translation_with_different_schemas(self):
        """Test LaTeX translation with different schema configurations."""
        schema1 = {"Users": ["id", "name", "email"]}
        schema2 = {
            "Products": ["product_id", "name", "price"],
            "Orders": ["order_id", "customer_id"],
        }

        # Test with schema1
        latex1 = self.rapt.to_qtree("pk_{id} Users;", schema1)
        assert "\\text{PK}(users) \\eq {id}" in latex1[0]

        # Test with schema2
        latex2 = self.rapt.to_qtree("inc=_{id, name} (Products, Orders);", schema2)
        assert "products[id] \\equiv orders[name]" in latex2[0]

    def test_latex_translation_error_handling(self):
        """Test LaTeX translation error handling."""
        # Test with invalid syntax
        with pytest.raises(Exception):
            self.rapt.to_qtree("invalid syntax", self.schema)

        # Test with empty input
        with pytest.raises(Exception):
            self.rapt.to_qtree("", self.schema)

    def test_latex_translation_consistency_across_runs(self):
        """Test that LaTeX translation is consistent across multiple runs."""
        test_case = "pk_{a, b} R; mvd_{x, y} S;"

        # Run multiple times
        latex1 = self.rapt.to_qtree(test_case, self.schema)
        latex2 = self.rapt.to_qtree(test_case, self.schema)

        assert latex1 == latex2
        assert len(latex1) == 2
        assert len(latex2) == 2


class TestDependencyLatexTranslationIntegration:
    """Integration tests for dependency LaTeX translation."""

    def setup_method(self):
        """Set up test fixtures."""
        self.grammar = DependencyGrammar()
        self.schema = {"R": ["a", "b", "c"], "S": ["x", "y"], "T": ["id", "name"]}
        self.builder = TreeBRD(self.grammar)

    def test_direct_grammar_latex_translation(self):
        """Test LaTeX translation using grammar directly."""
        from rapt2.transformers.qtree import qtree_translator

        # Parse dependency statements
        trees = self.builder.build("pk_{a, b} R;", self.schema)
        assert len(trees) == 1

        # Translate to LaTeX
        latex = qtree_translator.translate(list(trees))
        assert len(latex) == 1
        assert "\\text{PK}(r) \\eq {a,\\,b}" in latex[0]

    def test_multiple_dependency_types_latex_translation(self):
        """Test LaTeX translation of all dependency types in one go."""
        from rapt2.transformers.qtree import qtree_translator

        test_input = "pk_{a} R; mvd_{x, y} S; inc=_{a, b} (R, S); inc⊆_{x, y} (S, T);"
        trees = self.builder.build(test_input, self.schema)
        assert len(trees) == 4

        latex = qtree_translator.translate(list(trees))
        assert len(latex) == 4

        # Check each dependency type is properly translated
        expected_operators = [
            "\\text{PK}",
            "\\twoheadrightarrow",
            "\\equiv",
            "\\subseteq",
        ]
        for i, expected_op in enumerate(expected_operators):
            assert expected_op in latex[i]



from .core_grammar import CoreGrammar
from .dependency_grammar import DependencyGrammar
from .extended_grammar import ExtendedGrammar

GRAMMARS = {
    "Core Grammar": CoreGrammar,
    "Extended Grammar": ExtendedGrammar,
    "Dependency Grammar": DependencyGrammar,
}



from typing import List, Union

from ...treebrd.node import (AssignNode, CrossJoinNode, DefinitionNode, DifferenceNode,
                             FullOuterJoinNode, FunctionalDependencyNode,
                             InclusionEquivalenceNode,
                             InclusionSubsumptionNode, IntersectNode,
                             LeftOuterJoinNode, MultivaluedDependencyNode,
                             NaturalJoinNode, Node, PrimaryKeyNode,
                             ProjectNode, RelationNode, RenameNode,
                             RightOuterJoinNode, SelectNode, ThetaJoinNode,
                             UnionNode)
from ...treebrd.condition_node import (BinaryConditionNode, ConditionNode,
                                       IdentityConditionNode, UnaryConditionNode,
                                       UnaryConditionalOperator)
from ..base_translator import BaseTranslator
from .operators import latex_operator, conditional_latex_operator


class QTreeTranslator(BaseTranslator):
    """
    A Translator defining the operations for translating a relational algebra
    statement into a latex tree output.
    """

    def _get_latex_operator(self, operator):
        """
        Get the LaTeX operator for a given operator.
        For now, we use the hardcoded mapping, but this could be extended
        to use syntax-based mappings in the future.
        """
        return latex_operator[operator]

    def _get_conditional_latex_operator(self, operator):
        """
        Get the LaTeX operator for a given conditional operator.
        """
        return conditional_latex_operator[operator]

    def relation(self, node: RelationNode) -> str:
        """
        Translate a relation node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        return "[.${}$ ]".format(node.name.replace('_', r'\_') if node.name else '')

    def select(self, node: SelectNode) -> str:
        """
        Translate a select node into a latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        child = self.translate(node.child)
        return "[.${op}_{{{conditions}}}$ {child} ]".format(
            op=self._get_latex_operator(node.operator),
            conditions=self.translate_condition(node.conditions),
            child=child,
        )

    def project(self, node: ProjectNode) -> str:
        """
        Translate a project node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        child = self.translate(node.child)
        return "[.${op}_{{{attributes}}}$ {child} ]".format(
            op=self._get_latex_operator(node.operator),
            attributes=r",\,".join(node.attributes.names),
            child=child,
        )

    def rename(self, node: RenameNode) -> str:
        """
        Translate a rename node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        child = self.translate(node.child)
        attributes = ""
        if node.attributes:
            attributes = "({})".format(r",\,".join(node.attributes.names))
        return "[.${op}_{{{name}{attributes}}}$ {child} ]".format(
            op=self._get_latex_operator(node.operator),
            name=node.name.replace('_', r'\_') if node.name else '',
            attributes=attributes,
            child=child,
        )

    def assign(self, node: AssignNode) -> str:
        """
        Translate an assign node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        child = self.translate(node.child)
        attributes = ""
        if node.attributes:
            attributes = "({})".format(r",\,".join(node.attributes.names))
        return "[.${name}{attributes}$ {child} ]".format(
            name=node.name.replace('_', r'\_') if node.name else '', attributes=attributes, child=child
        )

    def definition(self, node: DefinitionNode) -> str:
        """
        Translate a definition node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        attributes = f"({r',\,'.join(node.attributes.names)})"

        return f"[.${(node.name.replace('_', r'\_') if node.name else '')}{attributes}$ ]"

    def theta_join(self, node: ThetaJoinNode) -> str:
        """
        Translate a join node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        return "[.${op}_{{{conditions}}}$ {left} {right} ]".format(
            op=self._get_latex_operator(node.operator),
            conditions=self.translate_condition(node.conditions),
            left=self.translate(node.left),
            right=self.translate(node.right),
        )

    def cross_join(self, node: CrossJoinNode) -> str:
        """
        Translate a cross node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        return self._binary(node)

    def natural_join(self, node: NaturalJoinNode) -> str:
        """
        Translate a natural join node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        return self._binary(node)

    def full_outer_join(self, node: FullOuterJoinNode) -> str:
        """
        Translate a full outer join node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        return "[.${op}_{{{conditions}}}$ {left} {right} ]".format(
            op=self._get_latex_operator(node.operator),
            conditions=self.translate_condition(node.conditions),
            left=self.translate(node.left),
            right=self.translate(node.right),
        )

    def left_outer_join(self, node: LeftOuterJoinNode) -> str:
        """
        Translate a left outer join node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        return "[.${op}_{{{conditions}}}$ {left} {right} ]".format(
            op=self._get_latex_operator(node.operator),
            conditions=self.translate_condition(node.conditions),
            left=self.translate(node.left),
            right=self.translate(node.right),
        )

    def right_outer_join(self, node: RightOuterJoinNode) -> str:
        """
        Translate a right outer join node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        return "[.${op}_{{{conditions}}}$ {left} {right} ]".format(
            op=self._get_latex_operator(node.operator),
            conditions=self.translate_condition(node.conditions),
            left=self.translate(node.left),
            right=self.translate(node.right),
        )

    def union(self, node: UnionNode) -> str:
        """
        Translate a union node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        return self._binary(node)

    def difference(self, node: DifferenceNode) -> str:
        """
        Translate a difference node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        return self._binary(node)

    def intersect(self, node: IntersectNode) -> str:
        """
        Translate an intersect node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        return self._binary(node)

    def _binary(
        self,
        node: Union[
            CrossJoinNode, NaturalJoinNode, UnionNode, DifferenceNode, IntersectNode
        ],
    ) -> str:
        """
        Translate a binary node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        return "[.${op}$ {left} {right} ]".format(
            op=self._get_latex_operator(node.operator),
            left=self.translate(node.left),
            right=self.translate(node.right),
        )

    def primary_key(self, node: PrimaryKeyNode) -> str:
        """
        Translate a primary key dependency node into a latex qtree node.
        :param node: a dependency node
        :return: a qtree subtree rooted at the node
        """
        attributes = r",\,".join(node.attributes)
        return f"[.${self._get_latex_operator(node.operator)}({node.relation_name}) \\eq {{{attributes}}}$ ]"

    def multivalued_dependency(self, node: MultivaluedDependencyNode) -> str:
        """
        Translate a multivalued dependency node into a latex qtree node.
        :param node: a dependency node
        :return: a qtree subtree rooted at the node
        """
        left_attr, right_attr = node.attributes
        op = self._get_latex_operator(node.operator)
        if isinstance(node.child, SelectNode):
            select_str = f"{self._get_latex_operator(node.child.operator)}_{{{self.translate_condition(node.child.conditions)}}} ({node.child.name})"
            return f"[.${select_str} : {left_attr} {op} {right_attr}$ ]"
        else:
            return f"[.${node.child.name} : {left_attr} {op} {right_attr}$ ]"

    def functional_dependency(self, node: FunctionalDependencyNode) -> str:
        """
        Translate a functional dependency node into a latex qtree node.
        :param node: a dependency node
        :return: a qtree subtree rooted at the node
        """
        left_attr, right_attr = node.attributes
        op = self._get_latex_operator(node.operator)
        if isinstance(node.child, SelectNode):
            select_str = f"{self._get_latex_operator(node.child.operator)}_{{{self.translate_condition(node.child.conditions)}}} ({node.child.name})"
            return f"[.${select_str} : {left_attr} {op} {right_attr}$ ]"
        else:
            return f"[.${node.child.name} : {left_attr} {op} {right_attr}$ ]"

    def inclusion_equivalence(self, node: InclusionEquivalenceNode) -> str:
        """
        Translate an inclusion equivalence dependency node into a latex qtree node.
        :param node: a dependency node
        :return: a qtree subtree rooted at the node
        """
        attr1, attr2 = node.attributes
        op = self._get_latex_operator(node.operator)

        # Handle left side
        if isinstance(node.left_child, SelectNode):
            left_str = f"{self._get_latex_operator(node.left_child.operator)}_{{{self.translate_condition(node.left_child.conditions)}}} ({node.left_child.name})"
        else:
            left_str = node.left_child.name

        # Handle right side
        if isinstance(node.right_child, SelectNode):
            right_str = f"{self._get_latex_operator(node.right_child.operator)}_{{{self.translate_condition(node.right_child.conditions)}}} ({node.right_child.name})"
        else:
            right_str = node.right_child.name

        return f"[.${left_str}[{attr1}] {op} {right_str}[{attr2}]$ ]"

    def inclusion_subsumption(self, node: InclusionSubsumptionNode) -> str:
        """
        Translate an inclusion subsumption dependency node into a latex qtree node.
        :param node: a dependency node
        :return: a qtree subtree rooted at the node
        """
        attr1, attr2 = node.attributes
        op = self._get_latex_operator(node.operator)

        # Handle left side
        if isinstance(node.left_child, SelectNode):
            left_str = f"{self._get_latex_operator(node.left_child.operator)}_{{{self.translate_condition(node.left_child.conditions)}}} ({node.left_child.name})"
        else:
            left_str = node.left_child.name

        # Handle right side
        if isinstance(node.right_child, SelectNode):
            right_str = f"{self._get_latex_operator(node.right_child.operator)}_{{{self.translate_condition(node.right_child.conditions)}}} ({node.right_child.name})"
        else:
            right_str = node.right_child.name

        return f"[.${left_str}[{attr1}] {op} {right_str}[{attr2}]$ ]"

    def identity_condition(self, node: IdentityConditionNode) -> str:
        """
        Translate an identity condition node into LaTeX.
        :param node: an identity condition node
        :return: LaTeX representation of the condition
        """
        return node.ident

    def unary_condition(self, node: UnaryConditionNode) -> str:
        """
        Translate a unary condition node into LaTeX.
        :param node: a unary condition node
        :return: LaTeX representation of the condition
        """
        op = self._get_conditional_latex_operator(node.op)
        if node.op == UnaryConditionalOperator.DEFINED:
            return f"{op}({self.translate_condition(node.child)})"
        else:
            return f"{op} {self.translate_condition(node.child)}"

    def binary_condition(self, node: BinaryConditionNode) -> str:
        """
        Translate a binary condition node into LaTeX.
        :param node: a binary condition node
        :return: LaTeX representation of the condition
        """
        op = self._get_conditional_latex_operator(node.op)
        return f"({self.translate_condition(node.left)} {op} {self.translate_condition(node.right)})"

    def translate_condition(self, condition: ConditionNode) -> str:
        """
        Translate a condition node into LaTeX.
        :param condition: a condition node
        :return: LaTeX representation of the condition
        """
        if isinstance(condition, IdentityConditionNode):
            return self.identity_condition(condition)
        elif isinstance(condition, UnaryConditionNode):
            return self.unary_condition(condition)
        elif isinstance(condition, BinaryConditionNode):
            return self.binary_condition(condition)
        else:
            raise ValueError(f"Unknown condition node type: {type(condition)}")


def translate(roots: List[Node], syntax=None) -> List[str]:
    """
    Translate a treebrd tree rooted at root into latex tree.
    :param roots: a list of treebrd nodes
    :param syntax: syntax instance for custom operators
    :return:  a string representing a latex qtree rooted at root
    """
    return [
        "\\Tree{root}".format(root=QTreeTranslator(syntax).translate(root)) for root in roots
    ]



import unittest

from pyparsing import ParseException

from rapt2.treebrd.grammars.condition_grammar import (ConditionGrammar,
                                                      get_attribute_references)
from tests.treebrd.grammars.grammar_test_case import GrammarTestCase


class TestConditionRules(GrammarTestCase):
    def setUp(self):
        self.parser = ConditionGrammar()

    def test_comparator(self):
        parse = self.parse_function(self.parser.comparator_op)

        self.assertEqual(parse("="), ["="])
        self.assertEqual(parse("!="), ["!="])
        self.assertEqual(parse("<>"), ["<>"])
        self.assertEqual(parse("<"), ["<"])
        self.assertEqual(parse("<="), ["<="])
        self.assertEqual(parse(">"), [">"])
        self.assertEqual(parse(">="), [">="])

    def test_comparator_exp(self):
        parse = self.parse_function(self.parser.comparator_op)

        self.assertRaises(ParseException, parse, "==")
        self.assertRaises(ParseException, parse, "><")
        self.assertRaises(ParseException, parse, "=<")
        self.assertRaises(ParseException, parse, "=>")

    def test_logical_unary(self):
        parse = self.parse_function(self.parser.not_op)

        self.assertEqual(parse("not"), ["not"])
        self.assertEqual(parse("NOT"), ["not"])

    def test_logical_binary(self):
        parse = self.parse_function(self.parser.logical_binary_op)

        self.assertEqual(parse("and"), ["and"])
        self.assertEqual(parse("AND"), ["and"])
        self.assertEqual(parse("or"), ["or"])
        self.assertEqual(parse("OR"), ["or"])

    def test_operand(self):
        parse = self.parse_function(self.parser.operand)

        self.assertEqual(parse('"Secret to life"'), ["'Secret to life'"])
        self.assertEqual(parse("is__"), ["is__"])
        self.assertEqual(parse("1.3337"), ["1.3337"])

    def test_operand_exp(self):
        parse = self.parse_function(self.parser.operand)

        self.assertRaises(ParseException, parse, "=")
        self.assertRaises(ParseException, parse, "Need space!")

    def test_condition(self):
        parse = self.parse_function(self.parser.condition)

        self.assertEqual(parse("2 > 7"), [["2", ">", "7"]])
        self.assertEqual(parse("secret = 3"), [["secret", "=", "3"]])
        self.assertEqual(parse('"name" != "mane"'), [["'name'", "!=", "'mane'"]])

    def test_condition_exp_fragments(self):
        parse = self.parse_function(self.parser.condition)

        self.assertRaises(ParseException, parse, "'Happiness' =")
        self.assertRaises(ParseException, parse, "< you")
        self.assertRaises(ParseException, parse, "<>")

    # Testing a single condition

    def test_conditions_single(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [["2", ">", "7"]]
        actual = parse("2 > 7")
        self.assertEqual(actual, expected)

    def test_conditions_par(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [["2", ">", "7"]]
        actual = parse("(2 > 7)")
        self.assertEqual(actual, expected)

    def test_conditions_not(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [["not", ["2", ">", "7"]]]
        actual = parse("not (2 > 7)")
        self.assertEqual(actual, expected)

    def test_conditions_and(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["2", ">", "7"], "and", ["3", "<>", "3"]]]
        actual = parse("2 > 7 and 3 <> 3")
        self.assertEqual(actual, expected)

    def test_conditions_and_par(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["2", ">", "7"], "and", ["3", "<>", "3"]]]
        actual = parse("(2 > 7) and (3 <> 3)")
        self.assertEqual(actual, expected)

    def test_conditions_or(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["2", ">", "7"], "or", ["3", "<>", "3"]]]
        actual = parse("(2 > 7) or (3 <> 3)")
        self.assertEqual(actual, expected)

    def test_conditions_unary_binary_simple(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["not", ["2", ">", "7"]], "and", ["3", "<>", "3"]]]
        actual = parse("not 2 > 7 and 3 <> 3")
        self.assertEqual(actual, expected)

    def test_conditions_unary_binary_simple_forced(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [["not", [["2", ">", "7"], "and", ["3", "<>", "3"]]]]
        actual = parse("not (2 > 7 and 3 <> 3)")
        self.assertEqual(actual, expected)

    def test_conditions_binary_unary_simple(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["3", "<>", "3"], "or", ["not", ["2", ">", "7"]]]]
        actual = parse("3 <> 3 or not 2 > 7")
        self.assertEqual(expected, actual)

    def test_conditions_chained_and_two(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["1", "=", "1"], "and", ["2", "=", "2"]]]
        actual = parse("1 = 1 and 2 = 2")
        self.assertEqual(actual, expected)

    def test_conditions_chained_and_three(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["1", "=", "1"], "and", [["2", "=", "2"], "and", ["3", "=", "3"]]]]
        actual = parse("1 = 1 and 2 = 2 and 3 = 3")
        self.assertEqual(actual, expected)

    def test_conditions_chained_and_four(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [
            [
                ["1", "=", "1"],
                "and",
                [["2", "=", "2"], "and", [["3", "=", "3"], "and", ["4", "=", "4"]]],
            ]
        ]
        actual = parse("1 = 1 and 2 = 2 and 3 = 3 and 4 = 4")
        self.assertEqual(actual, expected)

    def test_conditions_chained_or_two(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["1", "=", "1"], "or", ["2", "=", "2"]]]
        actual = parse("1 = 1 or 2 = 2")
        self.assertEqual(actual, expected)

    def test_conditions_chained_or_three(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["1", "=", "1"], "or", [["2", "=", "2"], "or", ["3", "=", "3"]]]]
        actual = parse("1 = 1 or 2 = 2 or 3 = 3")
        self.assertEqual(actual, expected)

    def test_conditions_chained_or_four(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [
            [
                ["1", "=", "1"],
                "or",
                [["2", "=", "2"], "or", [["3", "=", "3"], "or", ["4", "=", "4"]]],
            ]
        ]
        actual = parse("1 = 1 or 2 = 2 or 3 = 3 or 4 = 4")
        self.assertEqual(actual, expected)

    def test_conditions_mixed_chained_and_or(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [
            [
                ["1", "=", "1"],
                "and",
                [["2", "=", "2"], "or", [["3", "=", "3"], "and", ["4", "=", "4"]]],
            ]
        ]
        actual = parse("1 = 1 and 2 = 2 or 3 = 3 and 4 = 4")
        self.assertEqual(actual, expected)

    def test_conditions_mixed_chained_or_and(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [
            [
                ["1", "=", "1"],
                "or",
                [["2", "=", "2"], "and", [["3", "=", "3"], "or", ["4", "=", "4"]]],
            ]
        ]
        actual = parse("1 = 1 or 2 = 2 and 3 = 3 or 4 = 4")
        self.assertEqual(actual, expected)

    def test_conditions_chained_with_parentheses_and(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["1", "=", "1"], "and", [["2", "=", "2"], "and", ["3", "=", "3"]]]]
        actual = parse("1 = 1 and (2 = 2 and 3 = 3)")
        self.assertEqual(actual, expected)

    def test_conditions_chained_with_parentheses_or(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["1", "=", "1"], "or", [["2", "=", "2"], "or", ["3", "=", "3"]]]]
        actual = parse("1 = 1 or (2 = 2 or 3 = 3)")
        self.assertEqual(actual, expected)

    def test_conditions_chained_with_not_and(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["not", ["1", "=", "1"]], "and", ["not", ["2", "=", "2"]]]]
        actual = parse("not 1 = 1 and not 2 = 2")
        self.assertEqual(actual, expected)

    def test_conditions_chained_with_not_or(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["not", ["1", "=", "1"]], "or", ["not", ["2", "=", "2"]]]]
        actual = parse("not 1 = 1 or not 2 = 2")
        self.assertEqual(actual, expected)

    def test_conditions_chained_with_not_mixed(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [
            [
                ["not", ["1", "=", "1"]],
                "and",
                [
                    ["2", "=", "2"],
                    "or",
                    [["3", "=", "3"], "and", ["not", ["4", "=", "4"]]],
                ],
            ]
        ]
        actual = parse("not 1 = 1 and 2 = 2 or 3 = 3 and not 4 = 4")
        self.assertEqual(actual, expected)

    def test_conditions_chained_with_not_parentheses(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [
            [
                "not",
                [
                    ["1", "=", "1"],
                    "and",
                    [["2", "=", "2"], "or", [["3", "=", "3"], "and", ["4", "=", "4"]]],
                ],
            ]
        ]
        actual = parse("not (1 = 1 and 2 = 2 or 3 = 3 and 4 = 4)")
        self.assertEqual(actual, expected)

    def test_conditions_chained_complex_nested(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [
            [
                [["1", "=", "1"], "and", [["2", "=", "2"], "or", ["3", "=", "3"]]],
                "and",
                [["4", "=", "4"], "or", [["5", "=", "5"], "and", ["6", "=", "6"]]],
            ]
        ]
        actual = parse("(1 = 1 and (2 = 2 or 3 = 3)) and (4 = 4 or (5 = 5 and 6 = 6))")
        self.assertEqual(actual, expected)

    def test_conditions_chained_with_strings(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [
            [
                ["'name'", "=", "'john'"],
                "and",
                [
                    ["'age'", ">", "18"],
                    "or",
                    [["'status'", "=", "'active'"], "and", ["'role'", "!=", "'admin'"]],
                ],
            ]
        ]
        actual = parse(
            "'name' = 'john' and 'age' > 18 or 'status' = 'active' and 'role' != 'admin'"
        )
        self.assertEqual(actual, expected)

    def test_conditions_chained_with_numbers(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [
            [
                ["1.5", ">", "1.0"],
                "and",
                [
                    ["2.7", "<", "3.0"],
                    "or",
                    [["4.2", ">=", "4.0"], "and", ["5.1", "<=", "6.0"]],
                ],
            ]
        ]
        actual = parse("1.5 > 1.0 and 2.7 < 3.0 or 4.2 >= 4.0 and 5.1 <= 6.0")
        self.assertEqual(actual, expected)

    def test_conditions_chained_with_identifiers(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [
            [
                ["x", "=", "y"],
                "and",
                [["a", "!=", "b"], "or", [["c", "<", "d"], "and", ["e", ">=", "f"]]],
            ]
        ]
        actual = parse("x = y and a != b or c < d and e >= f")
        self.assertEqual(actual, expected)


class TestConditionAttrs(unittest.TestCase):
    def test_conditions_binary_unary(self):
        expected = ["answer"]
        self.assertEqual(expected, get_attribute_references("answer=42"))

    def test_conditions_binary_unary_with_prefix(self):
        expected = ["a.answer"]
        self.assertEqual(expected, get_attribute_references("a.answer=42"))

    def test_conditions_binary_unary_reversed(self):
        expected = ["answer"]
        self.assertEqual(expected, get_attribute_references("42=answer"))

    def test_conditions_binary_unary_no_attr(self):
        self.assertEqual([], get_attribute_references("42=42"))

    def test_conditions_binary_unary_two_attrs(self):
        expected = ["answer", "known"]
        self.assertEqual(expected, get_attribute_references("answer=known"))

    def test_conditions_binary_two_attrs(self):
        expected = ["answer", "known"]
        self.assertEqual(expected, get_attribute_references("answer=known and 42=42"))

    def test_conditions_binary_man_attrs(self):
        expected = ["answer", "known", "answer", "unknown"]
        self.assertEqual(
            expected, get_attribute_references("answer=known and answer=unknown")
        )



import unittest

from pyparsing import ParseException

from rapt2.treebrd.grammars.core_grammar import CoreGrammar


class TestSelect(unittest.TestCase):
    def setUp(self):
        self.parser = CoreGrammar()

    def test_single_condition(self):
        expression = "\\select_{age < 42} astronauts;"
        expected = [["\\select", [["age", "<", "42"]], ["astronauts"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_single_condition_reversed(self):
        expression = "\\select_{42 < age} astronauts;"
        expected = [["\\select", [["42", "<", "age"]], ["astronauts"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_single_condition_spacing(self):
        expression = "\\select_{age<42} astronauts;"
        expected = [["\\select", [["age", "<", "42"]], ["astronauts"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_single_condition_literal(self):
        expression = "\\select_{name = 'Douglas'} astronauts;"
        expected = [["\\select", [["name", "=", "'Douglas'"]], ["astronauts"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_condition_not(self):
        expression = "\\select_{not(age < 42)} astronauts;"
        expected = [["\\select", [["not", ["age", "<", "42"]]], ["astronauts"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_two_conditions_and(self):
        expression = "\\select_{age < 42 and flights > 1} astronauts;"
        expected = [
            [
                "\\select",
                [[["age", "<", "42"], "and", ["flights", ">", "1"]]],
                ["astronauts"],
            ]
        ]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_two_conditions_or(self):
        expression = "\\select_{age < 42 or name = 5} astronauts;"
        expected = [
            [
                "\\select",
                [[["age", "<", "42"], "or", ["name", "=", "5"]]],
                ["astronauts"],
            ]
        ]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    # Exceptions

    def test_exp_basic(self):
        expression = "\\select;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )

    def test_exp_no_relation(self):
        expression = "\\select_{age < 42};"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )

    def test_exp_no_attributes(self):
        expression = "\\select astronauts;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )

    def test_exp_empty_attribute_list(self):
        expression = "\\select_{} astronauts;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )

    def test_exp_space_before_attributes(self):
        expression = "\\select _{age < 42} astronauts;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )



import functools
from unittest import TestCase


class TestTransformer(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.schema = {
            "alpha": ["a1", "a2", "a3"],
            "alphatwin": ["a1", "a2", "a3"],
            "alphaprime": ["a1", "a4", "a5"],
            "beta": ["b1", "b2", "b3"],
            "gamma": ["g1", "g2"],
            "gammaprime": ["g1", "g2"],
            "gammatwin": ["g1", "g2"],
            "delta": ["d1", "d2"],
            "ambiguous": ["a", "a", "b"],
        }

    def translate_func(self, func, schema=None):
        schema = schema or self.schema
        return functools.partial(func, schema=schema)



import pytest

from src.rapt2.treebrd.grammars.dependency_grammar import DependencyGrammar
from src.rapt2.treebrd.node import (FunctionalDependencyNode,
                                    InclusionEquivalenceNode,
                                    InclusionSubsumptionNode,
                                    MultivaluedDependencyNode, Operator,
                                    PrimaryKeyNode, RelationNode, SelectNode)
from src.rapt2.treebrd.treebrd import TreeBRD


class TestDependencyGrammar:
    """Test cases for the DependencyGrammar class."""

    def setup_method(self):
        """Set up test fixtures."""
        self.grammar = DependencyGrammar()
        self.schema = {"R": ["a", "b", "c"], "S": ["x", "y"], "T": ["id", "name"]}

    def test_primary_key_dependency_parsing(self):
        """Test parsing of primary key dependencies."""
        # Test single attribute primary key
        result = self.grammar.parse("pk_{a} R;")
        assert len(result) == 1
        assert result[0][0] == "pk"
        assert list(result[0][1]) == ["a"]
        assert result[0][2] == "r"

        # Test multiple attributes primary key
        result = self.grammar.parse("pk_{a, b} R;")
        assert len(result) == 1
        assert result[0][0] == "pk"
        assert list(result[0][1]) == ["a", "b"]
        assert result[0][2] == "r"

    def test_multivalued_dependency_parsing(self):
        """Test parsing of multivalued dependencies."""
        # Test simple multivalued dependency (without conditions)
        result = self.grammar.parse("mvd_{a, b} R;")
        assert len(result) == 1
        assert result[0][0] == "mvd"
        assert list(result[0][1]) == ["a", "b"]
        assert result[0][2] == "r"

        # Test multivalued dependency with conditions
        result = self.grammar.parse("mvd_{a, b} \\select_{a = 1} R;")
        assert len(result) == 1
        assert result[0][0] == "mvd"
        assert list(result[0][1]) == ["a", "b"]
        assert result[0][2] == "\\select"
        assert list(result[0][3][0]) == ["a", "=", "1"]
        assert result[0][4] == "r"

        # Test complex multivalued dependency with conditions
        result = self.grammar.parse("mvd_{a, b} \\select_{a = 1 and b > 0} R;")
        assert len(result) == 1
        assert result[0][0] == "mvd"
        assert list(result[0][1]) == ["a", "b"]
        assert result[0][2] == "\\select"
        assert result[0][4] == "r"

    def test_functional_dependency_parsing(self):
        """Test parsing of functional dependencies."""
        # Test simple functional dependency (without conditions)
        result = self.grammar.parse("fd_{a, b} R;")
        assert len(result) == 1
        assert result[0][0] == "fd"
        assert list(result[0][1]) == ["a", "b"]
        assert result[0][2] == "r"

        # Test functional dependency with conditions
        result = self.grammar.parse("fd_{a, b} \\select_{a = 1} R;")
        assert len(result) == 1
        assert result[0][0] == "fd"
        assert list(result[0][1]) == ["a", "b"]
        assert result[0][2] == "\\select"
        assert list(result[0][3][0]) == ["a", "=", "1"]
        assert result[0][4] == "r"

        # Test complex functional dependency with conditions
        result = self.grammar.parse("fd_{a, b} \\select_{a = 1 and b > 0} R;")
        assert len(result) == 1
        assert result[0][0] == "fd"
        assert list(result[0][1]) == ["a", "b"]
        assert result[0][2] == "\\select"
        assert result[0][4] == "r"

    def test_simple_functional_dependency_parsing(self):
        """Test parsing of simple functional dependencies without conditions."""
        # Test simple functional dependency without conditions
        result = self.grammar.parse("fd_{a, b} R;")
        assert len(result) == 1
        assert result[0][0] == "fd"
        assert list(result[0][1]) == ["a", "b"]
        assert result[0][2] == "r"

    def test_inclusion_equivalence_parsing(self):
        """Test parsing of inclusion equivalence dependencies."""
        # Test simple inclusion equivalence
        result = self.grammar.parse("inc=_{a, b} (R, S);")
        assert len(result) == 1
        assert result[0][0] == "inc="
        assert list(result[0][1]) == ["a", "b"]
        assert len(result[0][2]) == 2
        assert result[0][2][0][0] == "r"
        assert result[0][2][1][0] == "s"

        # Test inclusion equivalence with select on left
        result = self.grammar.parse("inc=_{a, b} (\\select_{a = 1} R, S);")
        assert len(result) == 1
        assert result[0][0] == "inc="
        assert list(result[0][1]) == ["a", "b"]
        assert len(result[0][2]) == 2
        assert result[0][2][0][0] == "\\select"
        assert result[0][2][1][0] == "s"

        # Test inclusion equivalence with select on both sides
        result = self.grammar.parse("inc=_{a, b} (\\select_{a = 1} R, \\select_{x > 0} S);")
        assert len(result) == 1
        assert result[0][0] == "inc="
        assert list(result[0][1]) == ["a", "b"]
        assert len(result[0][2]) == 2
        assert result[0][2][0][0] == "\\select"
        assert result[0][2][1][0] == "\\select"

    def test_inclusion_subsumption_parsing(self):
        """Test parsing of inclusion subsumption dependencies."""
        # Test simple inclusion subsumption
        result = self.grammar.parse("inc⊆_{a, b} (R, S);")
        assert len(result) == 1
        assert result[0][0] == "inc⊆"
        assert list(result[0][1]) == ["a", "b"]
        assert len(result[0][2]) == 2
        assert result[0][2][0][0] == "r"
        assert result[0][2][1][0] == "s"

        # Test inclusion subsumption with select on left
        result = self.grammar.parse("inc⊆_{a, b} (\\select_{a = 1} R, S);")
        assert len(result) == 1
        assert result[0][0] == "inc⊆"
        assert list(result[0][1]) == ["a", "b"]
        assert len(result[0][2]) == 2
        assert result[0][2][0][0] == "\\select"
        assert result[0][2][1][0] == "s"

    def test_multiple_statements_parsing(self):
        """Test parsing of multiple dependency statements."""
        result = self.grammar.parse(
            "pk_{a} R; mvd_{b, c} \\select_{b > 0} S; fd_{x, y} \\select_{x > 0} T;"
        )
        assert len(result) == 3

        # Check first statement (primary key)
        assert result[0][0] == "pk"
        assert list(result[0][1]) == ["a"]
        assert result[0][2] == "r"

        # Check second statement (multivalued dependency with select)
        assert result[1][0] == "mvd"
        assert list(result[1][1]) == ["b", "c"]
        assert result[1][2] == "\\select"
        assert result[1][4] == "s"

        # Check third statement (functional dependency)
        assert result[2][0] == "fd"
        assert list(result[2][1]) == ["x", "y"]
        assert result[2][2] == "\\select"
        assert result[2][4] == "t"

    def test_invalid_syntax_raises_error(self):
        """Test that invalid syntax raises parsing errors."""
        with pytest.raises(Exception):
            self.grammar.parse("invalid syntax")

        with pytest.raises(Exception):
            self.grammar.parse("pk_{a} R")  # Missing semicolon

        with pytest.raises(Exception):
            self.grammar.parse("unknown_{a} R;")  # Unknown operator


class TestDependencyGrammarIntegration:
    """Test integration of DependencyGrammar with TreeBRD."""

    def setup_method(self):
        """Set up test fixtures."""
        self.grammar = DependencyGrammar()
        self.schema = {"R": ["a", "b", "c"], "S": ["x", "y"], "T": ["id", "name"]}
        self.builder = TreeBRD(self.grammar)

    def test_primary_key_node_creation(self):
        """Test creation of PrimaryKeyNode from dependency statements."""
        trees = self.builder.build("pk_{a, b} R;", self.schema)
        assert len(trees) == 1

        node = trees[0]
        assert isinstance(node, PrimaryKeyNode)
        assert node.operator == Operator.primary_key
        assert node.relation_name == "r"
        assert list(node.attributes) == ["a", "b"]

    def test_multivalued_dependency_node_creation(self):
        """Test creation of MultivaluedDependencyNode from dependency statements."""
        # Test simple multivalued dependency (without conditions)
        trees = self.builder.build("mvd_{a, b} R;", self.schema)
        assert len(trees) == 1

        node = trees[0]
        assert isinstance(node, MultivaluedDependencyNode)
        assert node.operator == Operator.multivalued_dependency
        assert node.relation_name == "r"
        assert list(node.attributes) == ["a", "b"]
        assert node.child is not None
        assert isinstance(node.child, RelationNode)

    def test_multivalued_dependency_node_creation_with_conditions(self):
        """Test creation of MultivaluedDependencyNode from dependency statements with conditions."""
        trees = self.builder.build("mvd_{a, b} \\select_{a = 1} R;", self.schema)
        assert len(trees) == 1

        node = trees[0]
        assert isinstance(node, MultivaluedDependencyNode)
        assert node.operator == Operator.multivalued_dependency
        assert node.relation_name == "r"
        assert list(node.attributes) == ["a", "b"]
        assert node.child is not None
        assert isinstance(node.child, SelectNode)

    def test_functional_dependency_node_creation_with_conditions(self):
        """Test creation of FunctionalDependencyNode from dependency statements with conditions."""
        trees = self.builder.build("fd_{a, b} \\select_{a = 1} R;", self.schema)
        assert len(trees) == 1

        node = trees[0]
        assert isinstance(node, FunctionalDependencyNode)
        assert node.operator == Operator.functional_dependency
        assert node.relation_name == "r"
        assert list(node.attributes) == ["a", "b"]
        assert node.child is not None
        assert isinstance(node.child, SelectNode)

    def test_functional_dependency_node_creation_simple(self):
        """Test creation of FunctionalDependencyNode from simple dependency statements."""
        trees = self.builder.build("fd_{a, b} R;", self.schema)
        assert len(trees) == 1

        node = trees[0]
        assert isinstance(node, FunctionalDependencyNode)
        assert node.operator == Operator.functional_dependency
        assert node.relation_name == "r"
        assert list(node.attributes) == ["a", "b"]
        assert node.child is not None
        assert isinstance(node.child, RelationNode)

    def test_inclusion_equivalence_node_creation(self):
        """Test creation of InclusionEquivalenceNode from dependency statements."""
        # Test simple inclusion equivalence
        trees = self.builder.build("inc=_{a, b} (R, S);", self.schema)
        assert len(trees) == 1

        node = trees[0]
        assert isinstance(node, InclusionEquivalenceNode)
        assert node.operator == Operator.inclusion_equivalence
        assert list(node.relation_names) == ["r", "s"]
        assert list(node.attributes) == ["a", "b"]
        assert isinstance(node.left_child, RelationNode)
        assert isinstance(node.right_child, RelationNode)

    def test_inclusion_equivalence_node_creation_with_select(self):
        """Test creation of InclusionEquivalenceNode with select clauses."""
        trees = self.builder.build("inc=_{a, b} (\\select_{a = 1} R, S);", self.schema)
        assert len(trees) == 1

        node = trees[0]
        assert isinstance(node, InclusionEquivalenceNode)
        assert node.operator == Operator.inclusion_equivalence
        assert list(node.relation_names) == ["r", "s"]
        assert list(node.attributes) == ["a", "b"]
        assert isinstance(node.left_child, SelectNode)
        assert isinstance(node.right_child, RelationNode)

    def test_inclusion_subsumption_node_creation(self):
        """Test creation of InclusionSubsumptionNode from dependency statements."""
        # Test simple inclusion subsumption
        trees = self.builder.build("inc⊆_{a, b} (R, S);", self.schema)
        assert len(trees) == 1

        node = trees[0]
        assert isinstance(node, InclusionSubsumptionNode)
        assert node.operator == Operator.inclusion_subsumption
        assert list(node.relation_names) == ["r", "s"]
        assert list(node.attributes) == ["a", "b"]
        assert isinstance(node.left_child, RelationNode)
        assert isinstance(node.right_child, RelationNode)

    def test_inclusion_subsumption_node_creation_with_select(self):
        """Test creation of InclusionSubsumptionNode with select clauses."""
        trees = self.builder.build("inc⊆_{a, b} (\\select_{a = 1} R, S);", self.schema)
        assert len(trees) == 1

        node = trees[0]
        assert isinstance(node, InclusionSubsumptionNode)
        assert node.operator == Operator.inclusion_subsumption
        assert list(node.relation_names) == ["r", "s"]
        assert list(node.attributes) == ["a", "b"]
        assert isinstance(node.left_child, SelectNode)
        assert isinstance(node.right_child, RelationNode)

    def test_multiple_dependency_statements(self):
        """Test parsing multiple dependency statements into multiple nodes."""
        trees = self.builder.build(
            "pk_{a} R; mvd_{x, y} \\select_{x > 0} S; fd_{id, name} \\select_{id > 0} T;", self.schema
        )
        assert len(trees) == 3

        # Check first node (primary key)
        assert isinstance(trees[0], PrimaryKeyNode)
        assert trees[0].relation_name == "r"

        # Check second node (multivalued dependency with select)
        assert isinstance(trees[1], MultivaluedDependencyNode)
        assert trees[1].relation_name == "s"
        assert isinstance(trees[1].child, SelectNode)

        # Check third node (functional dependency)
        assert isinstance(trees[2], FunctionalDependencyNode)
        assert trees[2].relation_name == "t"

    def test_dependency_node_equality(self):
        """Test equality comparison of dependency nodes."""
        trees1 = self.builder.build("pk_{a, b} R;", self.schema)
        trees2 = self.builder.build("pk_{a, b} R;", self.schema)

        # Test that nodes have the same properties
        node1 = trees1[0]
        node2 = trees2[0]
        assert node1.operator == node2.operator
        assert node1.relation_name == node2.relation_name
        assert list(node1.attributes) == list(node2.attributes)

        # Test inequality
        trees3 = self.builder.build("pk_{a, c} R;", self.schema)
        node3 = trees3[0]
        assert list(node1.attributes) != list(node3.attributes)

    def test_dependency_node_post_order(self):
        """Test post-order traversal of dependency nodes."""
        trees = self.builder.build("pk_{a, b} R;", self.schema)
        node = trees[0]

        post_order = node.post_order()
        assert len(post_order) == 1
        assert post_order[0] == node


class TestDependencyGrammarWithRapt:
    """Test DependencyGrammar integration with the main Rapt class."""

    def setup_method(self):
        """Set up test fixtures."""
        from src.rapt2.rapt import Rapt

        self.rapt = Rapt(grammar="Dependency Grammar")
        self.schema = {"R": ["a", "b", "c"], "S": ["x", "y"], "T": ["id", "name"]}

    def test_dependency_grammar_selection(self):
        """Test that DependencyGrammar can be selected via Rapt configuration."""
        assert self.rapt.builder.grammar.__class__.__name__ == "DependencyGrammar"

    def test_dependency_syntax_tree_generation(self):
        """Test generation of syntax trees for dependency statements."""
        trees = self.rapt.to_syntax_tree("pk_{a, b} R;", self.schema)
        assert len(trees) == 1
        assert isinstance(trees[0], PrimaryKeyNode)

    def test_multiple_dependency_syntax_trees(self):
        """Test generation of multiple syntax trees for multiple dependency statements."""
        trees = self.rapt.to_syntax_tree(
            "pk_{a} R; mvd_{x, y} \\select_{x > 0} S; fd_{id, name} \\select_{id > 0} T;", self.schema
        )
        assert len(trees) == 3
        assert isinstance(trees[0], PrimaryKeyNode)
        assert isinstance(trees[1], MultivaluedDependencyNode)
        assert isinstance(trees[1].child, SelectNode)
        assert isinstance(trees[2], FunctionalDependencyNode)



import functools

from rapt2.rapt import Rapt
from rapt2.transformers.sql import sql_translator
from rapt2.treebrd.grammars import CoreGrammar
from rapt2.treebrd.grammars.extended_grammar import ExtendedGrammar
from rapt2.treebrd.treebrd import TreeBRD
from tests.transformers.test_transfomer import TestTransformer


class TestSQLSequence(TestTransformer):
    grammar = CoreGrammar()

    def setUp(self):
        self.translate = self.translate_func(
            functools.partial(
                Rapt(grammar="Extended Grammar").to_sql_sequence, use_bag_semantics=True
            )
        )
        self.translate_set = self.translate_func(
            functools.partial(Rapt(grammar="Extended Grammar").to_sql_sequence)
        )


class TestRelation(TestSQLSequence):
    def test_single_relation(self):
        ra = "alpha;"
        expected = [["SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha"]]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_single_relation_set(self):
        ra = "alpha;"
        expected = [["SELECT DISTINCT alpha.a1, alpha.a2, alpha.a3 FROM alpha"]]
        actual = self.translate_set(ra)
        self.assertEqual(expected, actual)

    def test_multiple_relations(self):
        ra = "alpha; beta;"
        expected = [
            ["SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha"],
            ["SELECT beta.b1, beta.b2, beta.b3 FROM beta"],
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestSelect(TestSQLSequence):
    def test_single_select(self):
        ra = "\\select_{a1=a2} alpha;"
        expected = [
            [
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha",
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha WHERE (a1 = a2)",
            ]
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_multiple_select_with_multiple_conditions(self):
        ra = "\\select_{a1=2 or a1=1} \\select_{a2=2 or a2=1} alpha;"
        expected = [
            [
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha",
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha WHERE ((a2 = 2) OR (a2 = 1))",
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha "
                "WHERE (((a2 = 2) OR (a2 = 1))) AND (((a1 = 2) OR (a1 = 1)))",
            ]
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestProject(TestSQLSequence):
    def test_simple(self):
        ra = "\\project_{a1, a2} alpha;"
        expected = [
            [
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha",
                "SELECT alpha.a1, alpha.a2 FROM alpha",
            ]
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestRename(TestSQLSequence):
    def test_relation(self):
        ra = "\\rename_{apex} alpha;"
        expected = [
            [
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha",
                "SELECT apex.a1, apex.a2, apex.a3 "
                "FROM (SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) "
                "AS apex(a1, a2, a3)",
            ]
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_rename_attributes(self):
        ra = "\\rename_{(a, b, c)} alpha;"
        expected = [
            [
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha",
                "SELECT alpha.a, alpha.b, alpha.c "
                "FROM (SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) "
                "AS alpha(a, b, c)",
            ]
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_rename_all(self):
        ra = "\\rename_{apex(a, b, c)} alpha;"
        expected = [
            [
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha",
                "SELECT apex.a, apex.b, apex.c "
                "FROM (SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) "
                "AS apex(a, b, c)",
            ]
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestAssignment(TestSQLSequence):
    def test_relation(self):
        ra = "newalpha := alpha;"
        expected = [
            [
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha",
                "CREATE TEMPORARY TABLE newalpha(a1, a2, a3) AS "
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha",
            ]
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestJoin(TestSQLSequence):
    def test_relation(self):
        ra = "alpha \\join beta;"
        expected = [
            [
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha",
                "SELECT beta.b1, beta.b2, beta.b3 FROM beta",
                "SELECT alpha.a1, alpha.a2, alpha.a3, "
                "beta.b1, beta.b2, beta.b3 FROM "
                "(SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
                "CROSS JOIN "
                "(SELECT beta.b1, beta.b2, beta.b3 FROM beta) AS beta",
            ]
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_three_relations(self):
        self.maxDiff = None
        ra = "alpha \\join beta \\join gamma;"
        expected = [
            [
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha",
                "SELECT beta.b1, beta.b2, beta.b3 FROM beta",
                "SELECT alpha.a1, alpha.a2, alpha.a3, beta.b1, beta.b2, "
                "beta.b3 FROM "
                "(SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
                "CROSS JOIN "
                "(SELECT beta.b1, beta.b2, beta.b3 FROM beta) AS beta",
                "SELECT gamma.g1, gamma.g2 FROM gamma",
                "SELECT alpha.a1, alpha.a2, alpha.a3, beta.b1, beta.b2, "
                "beta.b3, gamma.g1, gamma.g2 FROM "
                "(SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
                "CROSS JOIN "
                "(SELECT beta.b1, beta.b2, beta.b3 FROM beta) AS beta "
                "CROSS JOIN "
                "(SELECT gamma.g1, gamma.g2 FROM gamma) AS gamma",
            ]
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestNaturalJoin(TestSQLSequence):
    grammar = ExtendedGrammar()

    def test_relation_simple(self):
        ra = "alpha \\natural_join alphatwin;"
        expected = [
            [
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha",
                "SELECT alphatwin.a1, alphatwin.a2, alphatwin.a3 FROM alphatwin",
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM "
                "(SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
                "NATURAL JOIN "
                "(SELECT alphatwin.a1, alphatwin.a2, alphatwin.a3 "
                "FROM alphatwin) AS alphatwin",
            ]
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestThetaJoin(TestSQLSequence):
    grammar = ExtendedGrammar()

    def test_relation(self):
        ra = "alpha \\join_{a1 = b1} beta;"
        expected = [
            [
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha",
                "SELECT beta.b1, beta.b2, beta.b3 FROM beta",
                "SELECT alpha.a1, alpha.a2, alpha.a3, "
                "beta.b1, beta.b2, beta.b3 FROM "
                "(SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
                "JOIN "
                "(SELECT beta.b1, beta.b2, beta.b3 FROM beta) AS beta "
                "ON (a1 = b1)",
            ]
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestSet:
    __test__ = False

    def test_simple(self):
        ra = "gamma {operator} gammatwin;".format(operator=self.ra_operator)

        root_list = TreeBRD(self.grammar).build(instring=ra, schema=self.schema)
        name = id(root_list[0])
        root_list = root_list[0].post_order()
        actual = sql_translator.translate(root_list, use_bag_semantics=True)

        expected = [
            "SELECT gamma.g1, gamma.g2 FROM gamma",
            "SELECT gammatwin.g1, gammatwin.g2 FROM gammatwin",
            "SELECT g1, g2 FROM ("
            "SELECT gamma.g1, gamma.g2 FROM gamma "
            "{operator} ALL "
            "SELECT gammatwin.g1, gammatwin.g2 FROM gammatwin) "
            "AS _{name}".format(operator=self.sql_operator, name=name),
        ]
        self.assertEqual(expected, actual)

    def test_simple_multiple(self):
        ra = "gamma {operator} gammatwin {operator} gammaprime;".format(
            operator=self.ra_operator
        )

        root_list = TreeBRD(self.grammar).build(instring=ra, schema=self.schema)
        root_name = id(root_list[0])
        child_name = id(root_list[0].left)
        root_list = root_list[0].post_order()
        actual = sql_translator.translate(root_list, use_bag_semantics=True)

        expected = [
            "SELECT gamma.g1, gamma.g2 FROM gamma",
            "SELECT gammatwin.g1, gammatwin.g2 FROM gammatwin",
            "SELECT g1, g2 FROM (SELECT gamma.g1, gamma.g2 FROM gamma "
            "{operator} ALL "
            "SELECT gammatwin.g1, gammatwin.g2 FROM gammatwin) AS _{name1}".format(
                operator=self.sql_operator,
                name1=child_name,
            ),
            "SELECT gammaprime.g1, gammaprime.g2 FROM gammaprime",
            "SELECT g1, g2 FROM ("
            "SELECT g1, g2 FROM "
            "(SELECT gamma.g1, gamma.g2 FROM gamma "
            "{operator} ALL "
            "SELECT gammatwin.g1, gammatwin.g2 FROM gammatwin) AS _{name1} "
            "{operator} ALL "
            "SELECT gammaprime.g1, gammaprime.g2 FROM gammaprime) AS _{name2}".format(
                operator=self.sql_operator, name1=child_name, name2=root_name
            ),
        ]
        self.assertEqual(expected, actual)


class TestUnion(TestSQLSequence, TestSet):
    ra_operator = "\\union"
    sql_operator = "UNION"


class TestDifference(TestSQLSequence, TestSet):
    ra_operator = "\\difference"
    sql_operator = "EXCEPT"


class TestIntersection(TestSQLSequence, TestSet):
    grammar = ExtendedGrammar()
    ra_operator = "\\intersect"
    sql_operator = "INTERSECT"


class TestDefined(TestSQLSequence):
    def test_simple_defined(self):
        ra = "\\select_{defined(a1)} alpha;"
        expected = [
            [
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha",
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha WHERE a1 IS NOT NULL",
            ]
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_not_defined(self):
        ra = "\\select_{not defined(a1)} alpha;"
        expected = [
            [
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha",
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha WHERE NOT a1 IS NOT NULL",
            ]
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_defined_and_defined(self):
        ra = "\\select_{defined(a1) and defined(a2)} alpha;"
        expected = [
            [
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha",
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha WHERE (a1 IS NOT NULL AND a2 IS NOT NULL)",
            ]
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_defined_or_defined(self):
        ra = "\\select_{defined(a1) or defined(a2)} alpha;"
        expected = [
            [
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha",
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha WHERE (a1 IS NOT NULL OR a2 IS NOT NULL)",
            ]
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_defined_with_equality(self):
        ra = "\\select_{defined(a1) and a1=5} alpha;"
        expected = [
            [
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha",
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha WHERE (a1 IS NOT NULL AND (a1 = 5))",
            ]
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_not_defined_with_equality(self):
        ra = "\\select_{not defined(a1) or a2=10} alpha;"
        expected = [
            [
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha",
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha WHERE (NOT a1 IS NOT NULL OR (a2 = 10))",
            ]
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_complex_defined_condition(self):
        ra = "\\select_{defined(a1) and (a2=5 or not defined(a3))} alpha;"
        expected = [
            [
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha",
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha WHERE (a1 IS NOT NULL AND ((a2 = 5) OR NOT a3 IS NOT NULL))",
            ]
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_defined_with_relation_attribute(self):
        ra = "\\select_{defined(alpha.a1)} alpha;"
        expected = [
            [
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha",
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha WHERE alpha.a1 IS NOT NULL",
            ]
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_multiple_defined_conditions(self):
        ra = "\\select_{defined(a1)} \\select_{defined(a2)} alpha;"
        expected = [
            [
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha",
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha WHERE a2 IS NOT NULL",
                "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha WHERE (a2 IS NOT NULL) AND (a1 IS NOT NULL)",
            ]
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)



from ..treebrd.grammars.syntax import Syntax
from ..treebrd.node import Operator


class BaseTranslator:
    """
    A Translator defining the operations for translating a relational algebra
    statement into some output format.
    """

    def __init__(self, syntax: Syntax | None = None):
        self.syntax = syntax or Syntax()
        self._translate_functions = {
            Operator.relation: self.relation,
            Operator.select: self.select,
            Operator.project: self.project,
            Operator.rename: self.rename,
            Operator.assign: self.assign,
            Operator.definition: self.definition,
            Operator.cross_join: self.cross_join,
            Operator.natural_join: self.natural_join,
            Operator.theta_join: self.theta_join,
            Operator.full_outer_join: self.full_outer_join,
            Operator.left_outer_join: self.left_outer_join,
            Operator.right_outer_join: self.right_outer_join,
            Operator.union: self.union,
            Operator.difference: self.difference,
            Operator.intersect: self.intersect,
            Operator.primary_key: self.primary_key,
            Operator.multivalued_dependency: self.multivalued_dependency,
            Operator.functional_dependency: self.functional_dependency,
            Operator.inclusion_equivalence: self.inclusion_equivalence,
            Operator.inclusion_subsumption: self.inclusion_subsumption,
        }

    def translate(self, node):
        """
        Translate a node into some output format.
        :param node: a treebrd node
        :return: a node's translation to some format
        """
        _translate = self._translate_functions.get(node.operator)
        if _translate is None:
            raise NotImplementedError(f"No translator for operator: {node.operator}")
        return _translate(node)

    def relation(self, node):
        raise NotImplementedError("Must be implemented by translation modules.")

    def select(self, node):
        raise NotImplementedError("Must be implemented by translation modules.")

    def project(self, node):
        raise NotImplementedError("Must be implemented by translation modules.")

    def rename(self, node):
        raise NotImplementedError("Must be implemented by translation modules.")

    def assign(self, node):
        raise NotImplementedError("Must be implemented by translation modules.")

    def definition(self, node):
        raise NotImplementedError("Must be implemented by translation modules.")

    def cross_join(self, node):
        raise NotImplementedError("Must be implemented by translation modules.")

    def natural_join(self, node):
        raise NotImplementedError("Must be implemented by translation modules.")

    def theta_join(self, node):
        raise NotImplementedError("Must be implemented by translation modules.")

    def full_outer_join(self, node):
        raise NotImplementedError("Must be implemented by translation modules.")

    def left_outer_join(self, node):
        raise NotImplementedError("Must be implemented by translation modules.")

    def right_outer_join(self, node):
        raise NotImplementedError("Must be implemented by translation modules.")

    def union(self, node):
        raise NotImplementedError("Must be implemented by translation modules.")

    def difference(self, node):
        raise NotImplementedError("Must be implemented by translation modules.")

    def intersect(self, node):
        raise NotImplementedError("Must be implemented by translation modules.")

    def primary_key(self, node):
        raise NotImplementedError("Must be implemented by translation modules.")

    def multivalued_dependency(self, node):
        raise NotImplementedError("Must be implemented by translation modules.")

    def functional_dependency(self, node):
        raise NotImplementedError("Must be implemented by translation modules.")

    def inclusion_equivalence(self, node):
        raise NotImplementedError("Must be implemented by translation modules.")

    def inclusion_subsumption(self, node):
        raise NotImplementedError("Must be implemented by translation modules.")


def translate(roots):
    """
    Translate a list of relational algebra trees into some output format.
    :param roots: a list of tree roots
    :return:  a list of translations
    """
    raise NotImplementedError("Must be implemented by translation modules.")



import copy

from .errors import RelationReferenceError


class Schema:
    """
    A Schema is a description of relational data.
    """

    def __init__(self, definition):
        self._data = {}
        for name, attributes in definition.items():
            self._data[name.lower()] = [attr.lower() for attr in attributes]

    def __eq__(self, other):
        if type(self) is not type(other):
            return False
        if self.to_dict() != other.to_dict():
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def contains(self, name):
        """
        Return true if the Schema contains a relation with the specified name.
        :param name: A name of a relation.
        :return: True if the Schema contains a relation with the specified name.
        """
        return name in self._data

    def to_dict(self):
        """
        Return a dictionary containing the name-attribute pairs in this Schema.
        :return: A dictionary of name-attribute pairs.
        """
        return copy.deepcopy(self._data)

    def get_attributes(self, name):
        """
        Return the list of attributes associated with the specified relation.
        :param name: A name of a relation in the Schema.
        :return: A list of attributes.
        :raise RelationReferenceError: Raised if the name does not exist.
        """
        attributes = self._data.get(name, None)
        if not attributes:
            raise RelationReferenceError(
                "Relation '{name}' does not exist.".format(name=name)
            )
        return attributes[:]

    def add(self, name, attributes):
        """
        Add the relation to the Schema.
        :param name: The name of a relation.
        :param attributes: A list of attributes for the relation.
        :raise RelationReferenceError: Raised if the name already exists.
        """
        if name in self._data:
            raise RelationReferenceError(
                "Relation '{name}' already exists.".format(name=name)
            )
        self._data[name] = attributes[:]
